//
//***************************************************************************
// Name    : HIDDescriptor.bas                                              *
// Author  : David John Barker                                              *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 10/01/2007                                                     *
// Version : 1.0                                                            *
// Notes   : Default USB HID descriptor module                              *
//         : =================== DO NOT EDIT THIS FILE ===================  *
//         : Save a copy in your current USB project folder and edit. If    *
//         : you rename the NEW descriptor file, then don't forget to set   *
//         : the new filename at the top of your main program using...      *
//         : #option USB_DESCRIPTOR = "MyDescriptor.bas"                    *
//***************************************************************************
//
module HIDDescriptor

// import modules...
include "usbconfig.bas"
include "usbdefs.bas"

// Device Descriptor
public const device_dsc(USB_DEV_DSC_SIZE) as byte = (                                         'usbdsc.c
    sizeof(USB_DEV_DSC),            // bLength - size of this descriptor IN bytes
    DSC_DEV,                        // bDescriptorType - device descriptor type
    $00,                            // bcdUSB (low byte)  - USB Spec Release Number IN BCD format (2.0)
    $02,                            // bcdUSB (high byte) - USB Spec Release Number IN BCD format (2.0)
    $00,                            // bDeviceClass - class Code
    $00,                            // bDeviceSubClass - subclass code
    $00,                            // bDeviceProtocol - protocol code
    USB_EP0_BUFF_SIZE,              // bMaxPacketSize - max packet size for EP0
    USB_VID and $00FF,              // idVendor (low byte)
    USB_VID >> 8,                   // idVendor (high byte)
    USB_PID and $00FF,              // idProduct (low byte)
    USB_PID >> 8,                   // idProduct (high byte)
    USB_DEVICE_VERSION and $00FF,   // bcdDevice (low byte) - device release number IN BCD format
    USB_DEVICE_VERSION >> 8,        // bcdDevice (high byte) - device release number IN BCD format
    USB_MANUFACTURER_INDEX,         // iManufacturer - manufacturer string index
    USB_PRODUCT_INDEX,              // iProduct - product string index
    USB_SERIAL_INDEX,               // iSerialNumber - device serial number string index
    $01                             // bNumConfigurations - mumber of possible configurations
)

// A typical configuration descriptor consists of these components:
//     At least one configuration descriptor (USB_CFG_DSC)
//     One or more interface descriptors (USB_INTF_DSC)
//     One or more endpoint descriptors (USB_EP_DSC)
// In addition, there is usually a descriptor string that provides a
// plain text description of the device

// Configuration 1 Descriptor
const CFG01_SIZE = 41
public const cfg01(CFG01_SIZE) as byte = (                                              'usbdsc.c

    // configuration Descriptor (cd01)
    sizeof(USB_CFG_DSC),    // Size of this descriptor in bytes
    DSC_CFG,                // CONFIGURATION descriptor type
    CFG01_SIZE,             // sizeof(cfg01) - Total length of data for this cfg
    0,                      // HIGH BYTE
    1,                      // Number of interfaces in this cfg
    1,                      // Index value of this configuration
    0,                      // Configuration string index
    _DEFAULT or _RWU,       // Attributes, see usbdefs_std_dsc.h
    USB_BUS_POWER,          // Max power consumption (2X mA)

    // interface descriptor (i00a00)
    sizeof(USB_INTF_DSC),   // Size of this descriptor in bytes
    DSC_INTF,               // INTERFACE descriptor type
    0,                      // Interface Number
    0,                      // Alternate Setting Number
    2,                      // Number of endpoints in this intf
    HID_INTF,               // Class code
    0,                      // Subclass code
    0,                      // Protocol code
    0,                      // Interface string index

    // HID class-specific descriptor (hid_i00a00)
    sizeof(USB_HID_DSC),    // Size of this descriptor in bytes
    DSC_HID,                // HID descriptor type
    $10,                    // HID Spec Release Number in BCD format (Low)
    $01,                    // HID Spec Release Number in BCD format (High)
    $00,                    // Country Code (0x00 for Not supported)
    HID_NUM_OF_DSC,         // Number of class descriptors, see usbcfg.h
    DSC_RPT,                // Report descriptor type
    HID_RPT01_SIZE,         // sizeof(hid_rpt01) - Size of the report descriptor
    0,                      // HIGH BYTE

    // endpoint descriptor (IN) (ep01i_i00a00)
    sizeof(USB_EP_DSC),     // Size of this descriptor in bytes
    DSC_EP,                 // Endpoint descriptor type
    _EP01_IN,               // Endpoint 1 IN
    _INT,                   // Interrupt transfers
    HID_INT_IN_EP_SIZE,     // Maximum packet size
    $00,                    // high byte
    HID_EP_IN_POLLING_MS,   // Polling interval (milliseconds)

    // enpoint descriptor (OUT) (ep01o_i00a00)
    sizeof(USB_EP_DSC),     // Size of this descriptor in bytes
    DSC_EP,                 // Endpoint descriptor type
    _EP01_OUT,              // Endpoint 1 OUT
    _INT,                   // Interrupt transfers
    HID_INT_OUT_EP_SIZE,    // Maximum packet size
    $00,                    // high byte
    HID_EP_OUT_POLLING_MS   // Polling interval (milliseconds)
)

// language string - unicode format...
const sd000_SIZE = 2     // size of data (2 bytes)
public const sd000(sd000_SIZE+2) as byte = (                                               'usbdsc.c
    sd000_SIZE+2,    // Size of this descriptor in bytes
    DSC_STR,        // DescriptorType
    $09,            // language ID - low byte
    $04             // language ID - high byte
)

// manufacturer string - unicode format...
const sd001_SIZE = 9*2       // length of unicode string data * 2
public const sd001(sd001_SIZE+2) as byte = (                                              'usbdsc.c
    sd001_SIZE+2,    // Size of this descriptor in bytes
    DSC_STR,        // DescriptorType
    "M", 0, "e", 0, "c", 0, "a", 0, "n", 0, "i", 0, "q", 0, "u", 0, "e", 0
)

// product string - unicode format...
const sd002_SIZE = 25*2      // length of unicode string data * 2
public const sd002(sd002_SIZE+2) as byte = (                                              'usbdsc.c
    sd002_SIZE+2,     // Size of this descriptor in bytes
    DSC_STR,         // DescriptorType
    "S", 0, "w", 0, "o", 0, "r", 0, "d", 0, "f", 0, "i", 0, "s", 0, "h", 0, " ", 0,
    "B", 0, "A", 0, "S", 0, "I", 0, "C", 0, " ", 0,
    "-", 0, " ", 0, "U", 0, "S", 0, "B", 0, " ", 0,
    "H", 0, "I", 0, "D", 0
)

{
// product serial number...
#if USB_HAVE_SERIAL_STRING
const sd003_SIZE = 4*2       // length of unicode string data * 2
public const sd003(sd003_SIZE+2) as byte = (                                              'usbdsc.c
    sd003_SIZE+2,    // Size of this descriptor in bytes
    DSC_STR,        // DescriptorType
    "1", 0, ".", 0, "0", 0, "0", 0
)
#endif
}
// product serial number...
#if USB_HAVE_SERIAL_STRING
const sd003_SIZE = 5*2       // length of unicode string data * 2
public const sd003(sd003_SIZE+2) as byte = (                                              'usbdsc.c
    sd003_SIZE+2,    // Size of this descriptor in bytes
    DSC_STR,        // DescriptorType
    "S", 0, "N", 0, "0", 0, "0", 0, "1", 0
)
#endif

//
//   This example report descriptor for a "generic HID" defines one
//   report of each type. Each report contains two bytes of data with
//   a vendor-defined Usage.
//
//   HID_RPT01_SIZE is defined in usbcfg.h and must equal the number of bytes
//   in the report descriptor.
//
public const hid_rpt01(HID_RPT01_SIZE) as byte = (                              'usbdsc.c
    $06, $A0, $FF,                  // Usage page (vendor defined)
    $09, $01,                       // Usage ID (vendor defined)
    $A1, $01,                       // Collection (application)

    // The Input report
    $09, $03,                       // Usage ID - vendor defined
    $15, $00,                       // Logical Minimum (0)
    $26, $FF, $00,                  // Logical Maximum (255)
    $75, $08,                       // Report Size (8 bits)
    $95, HID_REPORT_BYTES_IN,       // Report Count (2 fields)
    $81, $02,                       // Input (Data, Variable, Absolute)

    // The Output report
    $09, $04,                       // Usage ID - vendor defined
    $15, $00,                       // Logical Minimum (0)
    $26, $FF, $00,                  // Logical Maximum (255)
    $75, $08,                       // Report Size (8 bits)
    $95, HID_REPORT_BYTES_OUT,      // Report Count (2 fields)
    $91, $02,                       // Output (Data, Variable, Absolute)

    // The Feature report
    $09, $05,                       // Usage ID - vendor defined
    $15, $00,                       // Logical Minimum (0)
    $26, $FF, $00,                  // Logical Maximum (255)
    $75, $08,                       // Report Size (8 bits)
    $95, $02,                       // Report Count (2 fields)
    $B1, $02,                       // Feature (Data, Variable, Absolute)
    $C0                             // end collection
)

