//
//***************************************************************************
// Name    : main_hid.bas                                                   *
// Author  : Jerry Messina                                                  *
// Notice  : Copyright (c) 20011                                            *
//         : All Rights Reserved                                            *
// Date    : 4/7/2011                                                       *
// Version : 1.1                                                            *
// Notes   : USB HID main application example                               *
//         :                                                                *
// Ver 1.1 : added 'SF startup initialization' section to ensure register   *
//         : settings at poweron when used with a bootloader                *
//***************************************************************************
//

//
//-----------------------------------------------------------------------------
// hardware platforms (uncomment one section/customize for your hardware)
//-----------------------------------------------------------------------------
//
{
// Sparkfun USB whacker hardware
device = 18F2553
clock = 48
#define USB_WHACKER
}

{
// Microchip LPC (LowPinCount) 14K50 USB development board
device = 18F14K50
clock = 48
#define LPC_DEVBD
}

{
// 46J50 eval board hardware (USB self-powered)
device = 18F46J50
clock = 48
#define EVAL_J50
}

'{
// 47J53
device = 18F47J53
clock = 48
#define EVAL_J50
'}

//
//-----------------------------------------------------------------------------
// config settings
//-----------------------------------------------------------------------------
//
#if defined(USB_WHACKER)
'18F2553
config
    FOSC = HSPLL_HS,         ' External 24MHz xtal
    PLLDIV = 6,
    CPUDIV = OSC1_PLL2,      ' 48MHz cpu clock
    USBDIV = 2,              ' 48MHz usb clock
    FCMEN = OFF,
    IESO = OFF,
    PWRT = ON,
    BOR = OFF,
    BORV = 2,
    VREGEN = ON,
    WDT = OFF,
    WDTPS = 2048,
    CCP2MX = OFF,
    PBADEN = OFF,
    LPT1OSC = OFF,
    MCLRE = ON,
    STVREN = ON,
    LVP = OFF,
    XINST = OFF,
    DEBUG = OFF
#endif  // USB_WHACKER


#if defined(LPC_DEVBD)
'18F14K50
config
    FOSC = HS,              ' External 12MHz xtal
    CPUDIV = NOCLKDIV,      ' 48MHz cpu clock
    USBDIV = OFF,
    PLLEN = ON,             ' PLL enabled
    PCLKEN = ON,
    IESO = OFF,
    FCMEN = OFF,
    PWRTEN = ON,
    BOREN = ON,
    BORV = 30,
    WDTEN = OFF,
    WDTPS = 2048,
    HFOFST = ON,
    MCLRE = OFF,            ' must be off to read switch S1
    STVREN = ON,
    LVP = OFF,
    BBSIZ = OFF,
    XINST = OFF,
    DEBUG = OFF
#endif      // LPC_DEVBD


#if defined(EVAL_J50)
'18F46J50
public config
   OSC = INTOSCPLLO,                  ' internal osc 8MHz
   PLLDIV = 2,                        ' usb 48MHz
   CPUDIV = OSC1,                     ' 48MHz cpu clock
   FCMEN = OFF,
   IESO = OFF,
   WDTEN = OFF,
   WDTPS = 2048,
   STVREN = ON,
   XINST = OFF
'   DEBUG = OFF
#endif

//
//-----------------------------------------------------------------------------
// enable display of status messages to the Results window
//-----------------------------------------------------------------------------
//
// set false to disable messages
#option SHOW_BUILD_MSGS = true          // for this module
#option SHOW_USB_BUILD_MSGS = true      // for usb modules (off by default)

//
//-----------------------------------------------------------------------------
// ICD debugger (optional)
//-----------------------------------------------------------------------------
//
#option MPLAB_ICD = false
#if (MPLAB_ICD)
  include "mplabicd.bas"
#endif

// enable SF stack frame variable debugging
#option _showvar = true

//
//-----------------------------------------------------------------------------
// program build options (ORG_RESET and/or ORG_PROGRAM)
//-----------------------------------------------------------------------------
//
// build for bootloader/standalone operation
#variable BUILD_FOR_BOOTLOADER = true
#variable APP_PROGRAM_START = $001000       // must agree with bootloader setting

// if using the ICD, you don't want to use the bootloader
#if (MPLAB_ICD = true)
  #variable BUILD_FOR_BOOTLOADER = false
#endif

#if (BUILD_FOR_BOOTLOADER = true)           // build for use with bootloader
  #option org_reset = APP_PROGRAM_START
  #if (SHOW_BUILD_MSGS)
    #warning "building for use with bootloader"
  #endif
#else                                       // build for standalone use (no bootloader)
  #variable APP_PROGRAM_START = $000000     // normal location for startup
  #option org_program = APP_PROGRAM_START   // (can use this to offset location to match bootloader use)
#endif

//
//-----------------------------------------------------------------------------
// SF startup initialization
//-----------------------------------------------------------------------------
// note: include system.bas first before any other files
// this is VERY important if you're using a bootloader (requires system.bas V1.4 or later)
// be sure to set any system.bas WDT options prior to including the file
'#option WDT = true
include "system.bas"
#if not defined(SF_SYSTEM_VERSION) or (SF_SYSTEM_VERSION < 14)
  #if (BUILD_FOR_BOOTLOADER = true)
    #error "bootloader use requires system.bas v1.4 or later"
  #endif
#endif

//
//-----------------------------------------------------------------------------
// IPC module (optional)
// used for application->bootloader communications
// If used, this file MUST be included before any other libraries/startup code
// that might modify the RCON register
//-----------------------------------------------------------------------------
//
#option USE_IPC = true
#if (USE_IPC)
  include "ipcBootloader.bas"
#endif

//
//-----------------------------------------------------------------------------
// routines to test EEPROM and UserID bootloader functionality (optional)
// used to set data into the output HEX file
// note: these are not supported on J series devices
//-----------------------------------------------------------------------------
//
// UserID init
include "userid.bas"

// dummy eeprom data
'eeprom = ($55,1,2,3,4,5,6,7,8,$AA)

//
//-----------------------------------------------------------------------------
// User hardware routines
//-----------------------------------------------------------------------------
//
include "userio.bas"

//
//-----------------------------------------------------------------------------
// USB options and includes
//-----------------------------------------------------------------------------
//
// enable use of serial number
#option USB_HAVE_SERIAL_STRING = true

// HID descriptor file
#option USB_DESCRIPTOR = "HIDDescriptor.bas"

// disable extended ram use
// note: this option will not enable extended ram, only disable it (see usbmem.bas)
'#option USB_DISABLE_EXTENDED_RAM = true

// interrupt/polled mode
// note: you do not have to change the source when switching between modes... calls to
// USBService() will be optimized out when compiling for interrupt mode (USB_SERVICE=true)
// this option is set true by default in USBConfig.bas
#option USB_SERVICE = true

// import modules...
include "usbhid.bas"
include "usbsystem.bas"        // for USBSoftDetach

//
//-----------------------------------------------------------------------------
// USB report structures
//-----------------------------------------------------------------------------
//
// TX report
structure TTXReport
    Message as string
end structure

// RX report
structure TRXReport
    Message as string
end structure

//
// The functions HID.ReadReport() and HID.WriteReport() require USB ExtendedRam.
// If ExtendedRam is not defined, this code replicates the functionality of
// ReadReport()/WriteReport() using a local buffer and ReadArray()/WriteArray()
// so that the source code here doesn't have to change whether extended ram is
// used or not. This code is purely optional, and is done for convienence
//
#if (USB_EXTENDED_RAM_SIZE = 0)
  // local report buffers
  dim TXReport as TTXReport
  dim RXReport as TRXReport

  inline sub ReadReport()
      HID.ReadArray(@RXReport, sizeof(RXReport))
  end sub

  inline sub WriteReport()
      HID.WriteArray(@TXReport, sizeof(TXReport))
  end sub
#else       // USB_EXTENDED_RAM_SIZE > 0
  // use absolute buffers defined in USBHID
  dim TXReport as TTXReport absolute TX_REPORT_RAM
  dim RXReport as TRXReport absolute RX_REPORT_RAM
#endif

//
//-----------------------------------------------------------------------------
// main entry point
//-----------------------------------------------------------------------------
//
main:
    // setup pic io
    UserInit()

  #if (USE_IPC)
    // set RCON register bits to reflect pon
    IPC.reset_rcon()
  #endif
    delayms(10)

    // wait to connect to USB...
    UserLedOn()             // turn on led to indicate we're here waiting to enumerate...
    repeat
        USBservice()        // if using interrupts, calls to USBService() will be ignored
    until (HID.Attached)
    UserLedOff()            // turn it off now that we're attached

    // main program loop
    // read report messages and send/echo response (can use MikroE USB HIDTerminal)
    while (true)

        if (DataAvailable()) then
            UserLedOn()             // flash led to indicate we've got a packet
            ReadReport()            // read input message
            if (RXReport.Message = "reset") then            // reboot
                TXReport.Message = "rebooting..." + #13 + #10
                WriteReport()           // send response
                WaitForTX()             // wait for the msg to tx before we disconnect
                USBSoftDetach()         // disable USB module
                delayms(1000)           // allow the host a chance to sense detach
                IPC.reset()             // reboot
            elseif (RXReport.Message = "loader") then       // run bootloader
                TXReport.Message = "running bootloader" + #13 + #10
                WriteReport()           // send response
                WaitForTX()             // wait for the msg to tx before we disconnect
                USBSoftDetach()         // disable USB module
                delayms(1000)           // allow the host a chance to sense detach
                IPC.run_bootloader()    // run bootloader
            else                                            // just echo the command
                TXReport.Message = "rx: " + RXReport.Message + #13 + #10
                WriteReport()           // send response (echo data)
            endif
            UserLedOff()
        endif

        // if the USB_SERVICE option has been disabled, you MUST poll the USB connection
        // every 1ms or so to keep it alive and handle pending transactions on the bus...
        USBservice()

    end while
